import math
import re
from datetime import datetime

from api.miscellaneous import *
from api.database import *


JOBS_MIN_PAGE_SIZE = 10
JOBS_MAX_PAGE_SIZE = 500

WORKER_ID_PATTERN = re.compile("[0-9a-zA-Z._-]+")


def _job_db_to_json(job_db: DbResultRow) -> dict:
    job_json = {
        "id": job_db["id"],
        "type": job_db["type"],
        "priority": job_db["priority"],
        "state": job_db["state"],
        "is_canceled": bool(job_db["canceled"]),
        "creation_time": job_db["time_created"].strftime(API_DATETIME_FORMAT),
        "data": job_db["data"],
        "status": job_db["status"],
    }
    if job_db["worker"] is not None:
        job_json["worker_id"] = job_db["worker"]
    if job_db["last_ping"] is not None:
        job_json["last_worker_ping"] = job_db["last_ping"].strftime(API_DATETIME_FORMAT)
    if job_db["time_scheduled"] is not None:
        job_json["start_time"] = job_db["time_scheduled"].strftime(API_DATETIME_FORMAT)
    if job_db["time_finished"] is not None:
        job_json["finish_time"] = job_db["time_finished"].strftime(API_DATETIME_FORMAT)
    return job_json


def get_jobs(
        entries_per_page: int,
        page: int,
        type: str or None,
        worker_id: str or None,
        state: str or None) -> tuple[int, JsonTypes]:
    if entries_per_page < JOBS_MIN_PAGE_SIZE or entries_per_page > JOBS_MAX_PAGE_SIZE or page < 0:
        raise ValueError("Invalid page size or page number")  # pragma: no cover
    if type is None:
        type = "%"
    if worker_id is None:
        worker_id = "%"
    if state is None:
        state = "%"
    
    if worker_id == "%":
        worker_condition = ""
        worker_condition_values = []
    else:
        worker_condition = "AND \"worker\" LIKE ? OR (\"worker\" IS NULL AND ? = 'none')"
        worker_condition_values = [worker_id, worker_id]
    
    row_count_set, entries_db = db_pool.execute_read_statements_in_transaction(
        (f"""
        SELECT COUNT(*) AS "count" FROM "jobs"
        WHERE "type" LIKE ?
            AND "state" LIKE ?
            {worker_condition}
        """, [type, state, *worker_condition_values]),
        (f"""
        SELECT * FROM "jobs"
        WHERE "type" LIKE ?
            AND "state" LIKE ?
            {worker_condition}
        ORDER BY "time_created" DESC, "id" DESC
        LIMIT {entries_per_page} OFFSET {entries_per_page * page}
        """, [type, state, *worker_condition_values])
    )
    
    return (math.ceil(row_count_set[0]["count"] / entries_per_page),
            list(map(_job_db_to_json, entries_db)))


_SQL_GET_WORKERS = PreparedStatement(f"""
SELECT *
FROM "worker"
ORDER BY "last_ping" DESC
""")


def get_workers() -> list[dict]:
    workers_db = db_pool.execute_read_statement_in_transaction(_SQL_GET_WORKERS)
    workers_json = []
    for worker_db in workers_db:
        workers_json.append({
            "hostname": worker_db["hostname"],
            "last_ping": worker_db["last_ping"].strftime(API_DATETIME_FORMAT)
        })
    return workers_json


_SQL_DELETE_READY_FAILED_JOB = PreparedStatement(f"""
UPDATE "jobs"
SET "state" = 'deleted'
WHERE "id" = ?
    AND ("state" = 'ready' OR "state" = 'failed')
""")
_SQL_RESTART_FAILED_JOB = PreparedStatement(f"""
UPDATE "jobs"
SET "state" = 'ready', "canceled" = false
WHERE "id" = ?
    AND "state" = 'failed'
""")
_SQL_CANCEL_RUNNING_JOB = PreparedStatement(f"""
UPDATE "jobs"
SET "canceled" = true
WHERE "id" = ?
    AND "state" = 'running'
""")
_SQL_COPY_FINISHED_DELETED_JOB = PreparedStatement(f"""
INSERT INTO "jobs" ("type", "priority", "queue", "state", "time_created", "data")
    SELECT "type", "priority", "queue", 'ready', ?, "data"
    FROM "jobs"
    WHERE "id" = ?
        AND ("state" = 'finished'
            OR "state" = 'deleted'
        )
""")


def delete_ready_failed_job(id: int) -> bool:
    affected_count = db_pool.execute_write_transaction(
        lambda trans: trans.execute_statement_full_result_and_commit(
            _SQL_DELETE_READY_FAILED_JOB, id)[DB_RESULT_AFFECTED_ROWS])
    return affected_count > 0


def restart_failed_job(id: int) -> bool:
    affected_count = db_pool.execute_write_transaction(
        lambda trans: trans.execute_statement_full_result_and_commit(
            _SQL_RESTART_FAILED_JOB, id)[DB_RESULT_AFFECTED_ROWS])
    return affected_count > 0


def cancel_running_job(id: int) -> bool:
    affected_count = db_pool.execute_write_transaction(
        lambda trans: trans.execute_statement_full_result_and_commit(
            _SQL_CANCEL_RUNNING_JOB, id)[DB_RESULT_AFFECTED_ROWS])
    return affected_count > 0


def copy_finished_deleted_job(id: int) -> bool:
    affected_count = db_pool.execute_write_transaction(
        lambda trans: trans.execute_statement_full_result_and_commit(
            _SQL_COPY_FINISHED_DELETED_JOB, datetime.now(), id)[DB_RESULT_AFFECTED_ROWS])
    return affected_count > 0
