from typing import Iterable

from flask.wrappers import Request

from api.miscellaneous.constants import *
from api.miscellaneous.errors import (ApiClientException,
                                      ERROR_REQUEST_INVALID_PARAMETER, ERROR_REQUEST_MISSING_PARAMETER,
                                      ERROR_MALFORMED_JSON)

JsonTypes = dict[str, "JsonTypes"] or list["JsonTypes"] or str or int or bool or None


class CJsonValue:
    
    def __init__(self, data, path: str = "$"):
        self._data = data
        self._path = path
    
    def raise_error(self, message: str):
        raise ApiClientException(ERROR_REQUEST_INVALID_PARAMETER(self._path, message))
    
    def equals_json(self, json: JsonTypes):
        return self._data == json
    
    def is_null(self):
        return self._data is None
    
    def as_object(self) -> "CJsonObject":
        self.__check_value_type(dict, "object")
        return CJsonObject(self._data, self._path)
    
    def as_array(self) -> "CJsonArray":
        self.__check_value_type(list, "array")
        return CJsonArray(self._data, self._path)
    
    def as_bool(self) -> bool:
        self.__check_value_type(bool, "boolean")
        return self._data
    
    def as_sint32(self) -> int:
        return self.as_int(MIN_VALUE_SINT32, MAX_VALUE_SINT32)
    
    def as_int(self, min_value: int, max_value: int) -> int:
        self.__check_value_type(int, "int")
        if self._data < min_value:
            self.raise_error(f"Value must not be smaller than {min_value}")
        if self._data > max_value:
            self.raise_error(f"Value must not be greater than {max_value}")
        return self._data
    
    def as_string(self, max_length: int, min_length: int = None) -> str:
        self.__check_value_type(str, "string")
        if min_length is not None and len(self._data) < min_length:
            self.raise_error(f"String must not be shorter than {min_length}")
        if len(self._data) > max_length:
            self.raise_error(f"String must not be longer than {max_length}")
        return self._data
    
    def __check_value_type(self, expected_type: type, type_name: str):
        if self._data is None or not isinstance(self._data, expected_type):
            self.raise_error(f"Expected {type_name}")
        return self._data


class CJsonObject(CJsonValue):
    
    def __init__(self, data: {}, path: str = "$"):
        super().__init__(data, path)
    
    def has(self, key: str):
        return key in self._data
    
    def keys(self):
        return self._data.keys()
    
    def raise_if_present(self, key: str):
        if self.has(key):
            self.get(key).raise_error("This value should not be present")
    
    def get(self, key: str, optional: bool = False):
        if optional and not self.has(key):
            return None
        if key not in self._data:
            raise ApiClientException(ERROR_REQUEST_MISSING_PARAMETER(key))
        return CJsonValue(self._data[key], f"{self._path}.{key}")
    
    def get_object(self, key: str) -> "CJsonObject":
        return self.get(key).as_object()
    
    def get_array(self, key: str) -> "CJsonArray":
        return self.get(key).as_array()
    
    def get_bool(self, key: str, default: bool or None = None) -> bool:
        if default is not None and not self.has(key):
            return default
        return self.get(key).as_bool()
    
    def get_sint32(self, key: str, default: int or None = None) -> int:
        if default is not None and not self.has(key):
            return default
        return self.get(key).as_sint32()
    
    def get_int(self, key: str, min_value: int, max_value: int, optional: bool = False) -> int or None:
        if optional and not self.has(key):
            return None
        return self.get(key).as_int(min_value, max_value)
    
    def get_string(self, key: str, max_length: int, min_length: int = None, optional: bool = False) -> str or None:
        if optional and not self.has(key):
            return None
        return self.get(key).as_string(max_length, min_length)


class _ArrayIterator:
    
    def __init__(self, array: "CJsonArray"):
        super().__init__()
        self._array = array
        self._i = 0
    
    def __next__(self):
        if self._i >= self._array.length():
            raise StopIteration
        ele = self._array.get(self._i)
        self._i += 1
        return ele


class CJsonArray(CJsonValue, Iterable[CJsonValue]):
    
    def __init__(self, data: [], path: str = "$"):
        super().__init__(data, path)
        
    def length(self) -> int:
        return len(self._data)
    
    def get(self, index: int):
        if index < 0:
            raise ValueError(f"Negative index provided: {index}")  # pragma: no cover
        if index >= self.length():
            raise ApiClientException(ERROR_REQUEST_MISSING_PARAMETER(f"{self._path}[{index}]"))
        return CJsonValue(self._data[index], f"{self._path}[{index}]")
    
    def get_object(self, index: int) -> "CJsonObject":
        return self.get(index).as_object()
    
    def get_array(self, index: int) -> "CJsonArray":
        return self.get(index).as_array()
    
    def get_bool(self, index: int) -> bool:
        return self.get(index).as_bool()
    
    def get_sint32(self, index: int) -> int:
        return self.get(index).as_sint32()
    
    def get_int(self, index: int, min_value: int, max_value: int) -> int:
        return self.get(index).as_int(min_value, max_value)
    
    def get_string(self, index: int, max_length: int, min_length: int = None) -> str:
        return self.get(index).as_string(max_length, min_length)
    
    def __iter__(self):
        return _ArrayIterator(self)


def get_client_json(request: Request) -> CJsonObject:
    json = request.get_json(force=False, silent=True, cache=False)
    if json is None:
        raise ApiClientException(ERROR_MALFORMED_JSON)
    return CJsonObject(json)
