from typing import TypeVar, Generic
from abc import ABC, ABCMeta, abstractmethod

from api.miscellaneous import *
from api.database import *
from api.objects.changelog import ChangelogEntrySubject


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class ObjectField(ChangelogEntrySubject, ABC):
    __metaclass__ = ABCMeta
    
    def __init__(self,
                 api_id: str,
                 include_in_config: bool,
                 type_api_id: str,
                 default_value_json: JsonTypes or tuple[None] or None):
        self.api_id = api_id
        self.include_in_config = include_in_config
        self.type_api_id = type_api_id
        self.default_value_json = default_value_json
        
        self.__creation_config = None
        self.__object_class = None
        self.__description = None
    
    def on_added_to_class(self, object_class: "ObjectClass"):
        if self.__object_class is not None:
            raise RuntimeError("Already added to a class")  # pragma: no cover
        if object_class is None:
            raise ValueError("Class cannot be none")  # pragma: no cover
        self.__object_class = object_class
    
    def get_object_class(self) -> "ObjectClass":
        if self.__object_class is None:
            raise RuntimeError("This field has not been added to a class yet")  # pragma: no cover
        return self.__object_class
    
    def get_description(self) -> JsonTypes:
        if self.__description is not None:
            return self.__description
        self.__description = {
            "id": self.api_id,
            "type": self.type_api_id
        }
        if self.default_value_json is not None:
            self.__description["default_value"] = self.default_value_json if self.default_value_json != (None, ) else None
        self._add_type_restrictions_to(self.__description)
        return self.__description
    
    def get_value_json(self, current_value_json: JsonTypes) -> JsonTypes or None:
        return {
            "description": self.get_description(),
            "value": current_value_json
        }
    
    @abstractmethod
    def _add_type_restrictions_to(self, config_entry):
        pass  # pragma: no cover


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class ObjectColumnField(ObjectField):
    
    def __init__(self,
                 api_id: str,
                 include_in_config: bool,
                 type_api_id: str,
                 db_column: str,
                 default_value_json: JsonTypes or tuple[None] or None = None):
        super().__init__(api_id, include_in_config, type_api_id, default_value_json)
        self.db_column = db_ensure_valid_id(db_column)
    
    @abstractmethod
    def _add_type_restrictions_to(self, config_entry):
        pass  # pragma: no cover
    
    @abstractmethod
    def get_current_value_from_db(self, object_db: DbResultRow):
        pass  # pragma: no cover
    
    @abstractmethod
    def do_value_update(self,
                        transaction: AbstractTransaction,
                        changelog_update: "ChangelogUpdate",
                        object_id: int,
                        current_object_db: DbResultRow,
                        expected_old_client_value: CJsonValue or None,
                        new_client_value: CJsonValue) -> tuple[str, DbValueType] or None:
        pass  # pragma: no cover
    
    @abstractmethod
    def do_value_creation(self,
                          transaction: AbstractTransaction,
                          changelog_update: "CreationChangelogUpdate",
                          new_client_value: CJsonValue) -> tuple[str, DbValueType]:
        pass  # pragma: no cover


InterimValueType = TypeVar("InterimValueType")


# Cannot import due to circular dependency
# noinspection PyUnresolvedReferences
class ObjectIndirectField(ObjectField, Generic[InterimValueType]):
    
    def __init__(self,
                 api_id: str,
                 include_in_config: bool,
                 type_api_id: str,
                 default_value_json: JsonTypes or tuple[None] or None = None):
        super().__init__(api_id, include_in_config, type_api_id, default_value_json)
    
    @abstractmethod
    def queue_get_current_statements(self, transaction: ReadTransaction, object_id: int) -> tuple[FilledStatement, ...]:
        pass  # pragma: no cover
    
    @abstractmethod
    def get_json_value_from_results(self,
                                    transaction: ReadTransaction,
                                    object_db: DbResultRow,
                                    statements: tuple[FilledStatement, ...]) -> JsonTypes:
        pass  # pragma: no cover
    
    @abstractmethod
    def validate_and_convert_client_value(self,
                                          transaction: AbstractTransaction,
                                          object_id: int or None,
                                          client_value: CJsonValue) -> InterimValueType:
        pass  # pragma: no cover
    
    @abstractmethod
    def queue_update(self,
                     transaction: WriteTransaction,
                     changelog_update: "ChangelogUpdate",
                     object_id: int,
                     expected_old_client_value: CJsonValue or None,
                     new_value: InterimValueType):
        pass  # pragma: no cover

    @abstractmethod
    def queue_creation(self,
                       transaction: WriteTransaction,
                       changelog_update: "CreationChangelogUpdate",
                       object_id: int,
                       new_value: InterimValueType):
        pass  # pragma: no cover
